      .586
      .model flat, stdcall
      option casemap :none

; #########################################################################

      include \masm32\include\windows.inc
      include \masm32\include\kernel32.inc
      include \masm32\include\user32.inc

      includelib \masm32\lib\kernel32.lib
      includelib \masm32\lib\user32.lib

.data
    align 16
      f          dd 1,0,0,0,0,0,0,0
      f2p        dd 65536   ; 10000h
      it         dd 16384   ; 2^14

      p          dd 1,0,0,0,0,0,0,0
      cnt        dd 0
      subtract   dd 0

      minus      dd 0
      fermat     dd 14
      hidd       dd 7fffffffh  ; high bit zero
      lodd       dw 1, 65535   ; low word = one
      ; lodd       dd ffff0001h  ; low word = one   why doesn't this work ??

      fmtRem     db "%10lu Remainder (low 32 bits)",0
      szRem      db "                                     ",0
      szFactor   db "Found a factor !!!!!",0
      szCaption  db "Fermat 14 factoring ",0

.code

 ;      Background information:
 ;      
 ;      Purpose: To find factors of Fermat 14.
 ;
 ;      Fermat 14, the smallest known composite Fermat number without the factor known,
 ;      is 2^(2^14) + 1.
 ;      Possible factors of Fermat 14 are k*2^(14+2) + 1,
 ;      With k a positive integer  k = 1 ... roughly 2^(8192 - 16)
 ;      The possible factors are of the form k*65536 + 1.
 ;
 ;      There is a conjecture that the smallest factor is a number of maximum 256 bits.
 ;      
 ;      This is a promising target for Assembly language :
 ;        the problem is well understood,  does a number divide another evenly ( remainder 0)
 ;        the numbers are easy to manipulate ( powers of 2 ),
 ;        plenty of room to optimize,
 ;        side effect some optimized 256 bit x86 Assembly routines ( should scale to bigger )
 ;        
 ;      The algorithm used is:
 ;      
 ;         start with a Trial Factor (multiple of 65536) + 1
 ;       while not ( Factor found )  
 ;         do a mod of 2^(16384) + 1 by TF
 ;           if remainder 0    
 ;             Factor found
 ;           else
 ;             TF = TF + 65536
 ;             
 ;         It is using a powermod with 1 added when done.
 ;         Since 2^16384 = 2*2*2*..*2*2 (16384 2's multiplied together)
 ;         instead of creating the full Fermat 14 then dividing by TF, 
 ;         the variable p holds the current result of 2's multiplied together
 ;         with TF subtracted each time the number, p, exceeds TF.
 ; 
 ;         When 16384 multiplies of 2 are done 1 is added,
 ;         if the result equals TF, then the TF is a factor of Fermat 14.
 ; 
 ; 
 ;         Example of power mod using Fermat 5 (smaller, only 32 iterations, known factors)
 ;           2^(2^5) + 1 = 2^32 + 1 = 4294967296 + 1 = 4294967297
 ;         Trial Factors
 ;           k*2^(5+2) + 1 = k*2^7 + 1 = k*128 + 1  with k = 1,2,3,4,5,6 : 129,257,385,513,641,767
 ; 
 ;         for example k = 5 using the convention iteration,value
 ;         1,2 2,4 3,8 4,16 5,32 6,64 7,128 8,256 9,512 10,383 11,125 12,250 13,500 14,359
 ;         15,77 16,154 17,308 18,616 19,591 20,541 21,441 22,241 23,482 24,323 25,5 26,10
 ;         27,20 28,40 29,80 30,160 31,320 32,640  
 ;         then add 1 so 641  which equals the TF
 ;         so with k = 5,  TF = 5*128 + 1 = 641,  F14 = 4294967297   F14 mod TF = 0
 ; 
 ;         The assembly code works for factors less than 64 bits.
 ;         The code tests the TF with k = 3,  3 * 2^16 + 1 = 196609

 ;
 ;   Delphi/Pascal code, number of bits for factor depends on the size of unsigned integers used.
 ;        
 ;   fermat := 14;
 ;   f2p    := 65536;   (*  factor increment =  2^(fermat+2) = 2^16 = 65536  *)
 ;   f      := 1;
 ;   it     := 16384;   (*  2^14 = 16384 iterations  for power mod *)
 ;
 ;                      (*  k := 1 *)
 ;     f := f + f2p;    (*  f = 2^16 + 1 = 65537 *)
 ;     p := 1;    
 ;     c := 0;
 ;     while (c < it) do begin   (*  for c := 1 to it do begin *)
 ;       p := p + p;             (*    p := p * 2;             *)
 ;       if (p > f) then         (*    if (p > f) then         *)
 ;         p := p - f;           (*      p := p - f;           *)
 ;       c := c + 1;             (*                            *)
 ;     end;                      (*  end;                      *)
 ;
 ;     p := p + 1;
 ;     if p == f then
 ;       writeln('Factor found')
 ;     else
 ;       writeln(p,' remainder');
 ;

start:
  ; -------------------------------------------
  ; f := f + f2p
  ; -------------------------------------------
     mov    eax, f           
     add    eax, f2p
     mov    f, eax
     .IF CARRY?
        mov    eax, f+4
        adc    eax, 0
        mov    f+4, eax
     .ENDIF

     mov    eax, f           
     add    eax, f2p
     mov    f, eax
     .IF CARRY?
        mov    eax, f+4
        adc    eax, 0
        mov    f+4, eax
     .ENDIF

     mov    eax, f           
     add    eax, f2p
     mov    f, eax
     .IF CARRY?
        mov    eax, f+4
        adc    eax, 0
        mov    f+4, eax
     .ENDIF

  ; -------------------------------------------
  ; cnt := 0
  ; -------------------------------------------
     mov    eax, 0
     mov    cnt, eax
     
  ; -------------------------------------------
  ; while cnt < it do
  ; -------------------------------------------
     .WHILE (eax < it)
     
  ; -------------------------------------------
  ; p := p * 2  EQUIV  p := p + p
  ; -------------------------------------------
        mov    eax, p
        shld   p+4, eax, 1
        shl    p, 1

;        mov    eax, p    ; addition version
;        add    eax, eax
;        mov    p, eax
;        mov    eax, p+4
;        adc    eax, eax
;        mov    p+4, eax

  ; -------------------------------------------
  ; if p > f then
  ; -------------------------------------------
        mov    subtract, 0
        mov    edx, p+4
        mov    eax, p
        .IF (edx > f+4)
           mov    subtract, 1
        .ELSEIF (edx == f+4) && (eax > f)
           mov    subtract, 1
        .ENDIF
        
  ; -------------------------------------------
  ; p := p - f
  ; -------------------------------------------
       .IF (subtract == 1)
         mov    eax, p
         sub    eax, f
         mov    p, eax
         mov    eax, p+4
         sbb    eax, f+4
         mov    p+4, eax
       .ENDIF

  ; -------------------------------------------
  ; c := c + 1
  ; -------------------------------------------
       mov    eax, cnt
       add    eax, 1
       mov    cnt, eax
     .ENDW

  ; -------------------------------------------
  ; p := p + 1
  ; -------------------------------------------
     add    p, 1
     adc    p+4, 0

  ; -------------------------------------------
  ; if (p == f) then
  ;   writeln('Factor')
  ; else
  ;   writeln(p,' Remainder');
  ; -------------------------------------------
     invoke wsprintf,ADDR szRem,ADDR fmtRem,p
     mov    edx, p+4
     mov    eax, p
     .IF (edx == f+4) && (eax == f)
       invoke MessageBox,NULL,ADDR szFactor,ADDR szCaption,MB_OK
     .ELSE
       invoke MessageBox,NULL,ADDR szRem,   ADDR szCaption,MB_OK
     .ENDIF
     
     mov    eax, 0
     invoke ExitProcess,0
end start